/**********************************************************************

  The generic part of solving proximity equations.

  Does not take into account the actual proximity relation.

  Should be included at the beginning of each file where a specific 
  proximity relation is defined.

  The initial version: 2023-03-21.
  Version 0.0.1.0. Last edit on 2023-04-30.
  (c) RISC, JKU Linz.

**********************************************************************/


%----------------------------------------------------------------------
%% solve(?S:term, ?T:term, +Threshold:float, -Actual:float).
%
%  Succeeds if S and T can be made proximal to each other with
%  degree Actual, which is greater than or equal to the given
%  Threshold.
%----------------------------------------------------------------------

solve(_, _, Threshold, _) :-
    var(Threshold),
    !,
    Lines=['The threshold should be a number within the real interval [0,1].'],
    print_message_lines(user_error, 'ERROR: ', Lines),
    fail.
solve(S, T, Threshold, Actual) :-
    nonvar(Threshold),
    prox(S, T, Actual),
    Threshold =< Actual.
solve(S, T, 0, Actual) :-
    \+ prox(S, T, Actual),
    Actual = 0,
    Lines = ['Approximation degree 0 means that for any instantiation of variables, the given terms ~w and ~w can not be made proximal with a positive degree.'-[S,T]],
    print_message_lines(user_error, '\nWARNING: ', Lines).


%----------------------------------------------------------------------
%% all_instances(?S:term, ?T:term, +Threshold:float, -Instances:list).
%
%  Succeeds if Instances is the list of pairs [inst,deg] where
%  inst is an instance of S such that inst is proximal to T with 
%  the proximity degree deg >= Threshold.
% 
%  Each instance of S that is at least Threshold-proximal to T
%  appears in Instances only once, with the maximal degree deg. 
%----------------------------------------------------------------------

all_instances(S, T, Threshold, Instances) :-
    ground(T),
    !,
    findall([S,Actual], solve(S, T, Threshold, Actual), L),
    sort(L, L_sorted),
    delete_duplicates_with_smaller_degree(L_sorted, Instances).

all_instances(S, T, Threshold, Instances) :-
    findall([S,T,Actual], solve(S, T, Threshold, Actual), L),
    sort(L, L_sorted),
    delete_duplicates_with_smaller_degree_ng(L_sorted, Instances).

delete_duplicates_with_smaller_degree([], []).
delete_duplicates_with_smaller_degree([[I,_],[I,D2]|T], Res) :-
    !,
    delete_duplicates_with_smaller_degree([[I,D2]|T], Res).
delete_duplicates_with_smaller_degree([X|Tin], [X|Tout]) :-
    delete_duplicates_with_smaller_degree(Tin, Tout).

delete_duplicates_with_smaller_degree_ng([], []).
delete_duplicates_with_smaller_degree_ng([[IS,IT,_],[IS,IT,D2]|T], Res) :-
    !,
    delete_duplicates_with_smaller_degree_ng([[IS,IT,D2]|T], Res).
delete_duplicates_with_smaller_degree_ng([X|Tin], [X|Tout]) :-
    delete_duplicates_with_smaller_degree_ng(Tin, Tout).


%----------------------------------------------------------------------
%% best_instances(?S:term, ?T:term, +Threshold:float, -Instances:list.
%
%  Succeeds if Instances is the list of pairs [inst,maxdeg] where
%  inst is an instance of S such that inst is proximal to T with 
%  the proximity degree maxdeg >= Threshold.
% 
%  All elements in Instances have the same degree maxdeg. No instance 
%  if S is close to T with degree higher than maxdeg
%----------------------------------------------------------------------

best_instances(S, T, Threshold, Instances) :-
    ground(T),
    !,
    findall([S,Actual], solve(S, T, Threshold, Actual), L),
    sort(2, @>=, L, L_sorted),
    return_instances_with_the_highest_degree(L_sorted, Instances).

best_instances(S, T, Threshold, Instances) :-
    findall([[S,T],Actual], solve(S, T, Threshold, Actual), L),
    sort(2, @>=, L, L_sorted),
    return_instances_with_the_highest_degree_ng(L_sorted, Instances),
    !.

return_instances_with_the_highest_degree([], []).
return_instances_with_the_highest_degree([X], [X]).
return_instances_with_the_highest_degree([[I,D1],[_,D2]|_], [[I,D1]]) :-
    D1 > D2,
    !.
return_instances_with_the_highest_degree([[I,D],[I,D]|Tin], [[I,D]|Tout]) :-
    !,
    return_instances_with_the_highest_degree(Tin, Tout).
return_instances_with_the_highest_degree([[I1,D],[I2,D]|Tin], [[I1,D]|Tout]) :-
    return_instances_with_the_highest_degree([[I2,D]|Tin], Tout).


return_instances_with_the_highest_degree_ng([], []).
return_instances_with_the_highest_degree_ng([[[IS,IT],D]], [[IS,IT,D]]).
return_instances_with_the_highest_degree_ng([[[IS,IT],D1],[[_,_],D2]|_], [[IS,IT,D1]]) :-
    D1 > D2,
    !.
return_instances_with_the_highest_degree_ng([[[IS,IT],D],[[IS,IT],D]|Tin], [[IS,IT,D]|Tout]) :-
    !,
    return_instances_with_the_highest_degree_ng(Tin, Tout).
return_instances_with_the_highest_degree_ng([[[IS1,IT1],D],[[IS2,IT2],D]|Tin], [[IS1,IT1,D]|Tout]) :-
    return_instances_with_the_highest_degree_ng([[[IS2,IT2],D]|Tin], Tout).



%----------------------------------------------------------------------
%% is_proximal(?S:term, ?T:term, ?Delta:float).
%
%  Succeeds if S and T are proximal to each other with degree Delta
%  (at least). When Delta is given (i.e., when the threshold is fixed),
%  it checks whether S and T are proximal with the degree at least
%  Delta. When Delta in not given, i.e., when it is a free variable,
%  is_proximal computes its actual value as the exact proximity degree
%  between S and T.
%
%  Unlike prox and solve, does not instantiate variables in S and T, 
%  treating them as unique constants.
%----------------------------------------------------------------------

is_proximal(S, T, Threshold) :-
    nonvar(Threshold),
    grounding_term([S,T], [S_gr,T_gr]),
    solve(S_gr, T_gr, Threshold, _),
    !.
is_proximal(S, T, Actual) :-
    var(Actual),
    grounding_term([S,T], [S_gr,T_gr]),
    solve(S_gr, T_gr, 0, Actual),
    !.
is_proximal(_, _, 0).


%% grounding_term(+T:term, -G:ground term).
%
%  Given a term T, computes G that is obtained from T by replacing
%  each variable with a unique (automatically generated) constant.
%  Is used in the definition of predicate is_proximal.
    
grounding_term(Var, Constant) :-
    var(Var),
    !,
    gensym(c, Constant).
grounding_term(Term, Ground_term) :-
    Term =.. [H|Arg_list],
    grounding_term_list(Arg_list, Arg_list_ground),
    Ground_term =.. [H|Arg_list_ground].


%% grounding_term_list(+TL:term list, -GL:ground term list).
%
%  Given a term list TL, computes the list GL that is obtained from
%  TL by replacing each variable with a unique (automatically generated)
%  constant, applying grounding_term to each element of TL.

grounding_term_list([], []).
grounding_term_list([Term|Terms], [Ground_term|Ground_terms]) :-
    grounding_term(Term, Ground_term),
    Term = Ground_term,
    grounding_term_list(Terms, Ground_terms).



%% prox_list(?L1:term list, ?L2:term list, +Acc:float, ?Delta:float).
%
%  It is an auxuliary predicate for the definition of prox below.
%  Applies prox pairwise to elements of the lists L1 and L2.
%  Acc is the accumulator for the proximity degree.
%  Delta is the final degree of proximity between L1 and L2.
%  The minimum T-norm is used to compute proximity degrees
%  between two compound objects (terms or lists of terms).
%
%  Remark:  prox_list is defined before prox, because the clauses
%  defining prox should be the last ones in this file.
%  The reason: this file is supposed to be included at the beginning
%  of the concrete relation files consisting of other prox clauses
%  and in this way, all the prox clauses will be groupped together.

prox_list([], [], Delta, Delta).
prox_list([H1|T1], [H2|T2], Delta_acc, Delta) :-
    prox(H1, H2, Delta_H),
    Delta_acc_new is min(Delta_acc, Delta_H),
    prox_list(T1, T2, Delta_acc_new, Delta).


%----------------------------------------------------------------------
%% prox(?S:term, ?T:term, ?Delta:float).
%
%  Succeeds if there are instatiations of variables in S, T, and Delta
%  such that the instances of S and T are proximal to each other
%  with the proximity degree Delta.
%
%  Delta is supposed to be positive.
%----------------------------------------------------------------------

prox(S, T, 1) :-
    S == T,
    !.
prox(S, T, 1) :-
    var(S),
    atomic(T),
    S = T.
%   !.
prox(S, T, 1) :-
    atomic(S),
    var(T),
    S = T.
prox(S, T, Delta) :-
    nonvar(S),
    nonvar(T),
    S =.. [F,AS|LS],
    T =.. [F,AT|LT],
    prox(AS, AT, Delta1),
    prox_list(LS, LT, Delta1, Delta).
prox(S, T, Delta) :-
    nonvar(S),
    var(T),
    S =.. [F,AS|LS],
    prox(AS, AT, Delta1),
    prox_list(LS, LT, Delta1, Delta),
    T =.. [F,AT|LT].
prox(S, T, Delta) :-
    var(S),
    nonvar(T),
    T =.. [F,AT|LT],
    prox(AS, AT, Delta1),
    prox_list(LS, LT, Delta1, Delta),
    S =.. [F,AS|LS].

