/***************************************************

 These clauses define a sample proximity relation.
 
 Relation R1, 2023-03-21.
 Last version: 2023-04-30.

***************************************************/

:-include('generic.pl').

% A proximity equation between two nonconstant terms.
% Both symmetric versions should be written,
% unless the symmetric version causes looping.

prox(f(X1,X2,X3,X4), g(Y1,Y2,_,Y4,Y5), Delta) :-
  % defining and degree constraints
    prox(X2, Y1, Delta1),
    (Delta1 >= 0.4),             
    prox(X4, Y2, Delta2),
    (Delta2 >= 0.5),
    prox(X4, Y4, Delta3),
    (Delta3 >= 0.6),
  % restricting and degree constraints        
    prox(X1, h(X2,b), 1),
    prox(X2, X3, Alpha1),
    (Alpha1 >= 0.3),
    (Alpha1 =< 0.7),
    prox(Y4, Y5, Beta1),
    (Beta1 >= 0.4),
  % computing the proximity degree.
    min_list([0.7,Delta1,Delta2,Delta3], Delta).
prox(g(Y1,Y2,Y3,Y4,Y5), f(X1,X2,X3,X4), Delta) :-
   prox(f(X1,X2,X3,X4), g(Y1,Y2,Y3,Y4,Y5), Delta).


% Proximity equatiosn between different constants.
% Both symmetric versions should be written.

prox(a, b, 0.6).
prox(b, a, 0.6).

prox(b, c, 0.7).
prox(c, b, 0.7).




/*****************************************

Sample goals (the predicate solve is define in generic.pl):

?- solve(X, a, 0, Deg).
X = a,
Deg = 1 ;
X = b,
Deg = 0.6 ;
false.

?- solve(X, a, 0.8, Deg).
X = a,
Deg = 1 ;
false.

?- solve(b, X, 0, Deg).
X = b,
Deg = 1 ;
X = a,
Deg = 0.6 ;
X = c,
Deg = 0.7 ;
false.

?- solve(b, X, 0.7, Deg).
X = b,
Deg = 1 ;
X = c,
Deg = 0.7 ;
false.

?- solve(h(X,X), h(a,c), 0, Deg).
X = b,
Deg = 0.6 ;
false.

?- solve(h(X,c), h(a,X), 0, Deg).
X = b,
Deg = 0.6 ;
false.

?- solve(h(X,X), h(b,b), 0, Deg).
X = b,
Deg = 1 ;
X = a,
Deg = 0.6 ;
X = c,
Deg = 0.7 ;
false.

?- solve(h(X,X), h(b,b), 0.7, Deg).
X = b,
Deg = 1 ;
X = c,
Deg = 0.7 ;
false.

?- solve(f(h(Z1,Z2),Z1,Z2,Z3), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, Deg).
Z1 = a,
Z2 = b,
Z3 = h(a,c),
Deg = 0.6 ;
Z1 = a,
Z2 = b,
Z3 = h(a,b),
Deg = 0.6 ;
Z1 = a,
Z2 = b,
Z3 = h(b,c),
Deg = 0.6 ;
Z1 = a,
Z2 = b,
Z3 = h(b,b),
Deg = 0.6 ;
Z1 = c,
Z2 = b,
Z3 = h(a,c),
Deg = 0.7 ;
Z1 = c,
Z2 = b,
Z3 = h(a,b),
Deg = 0.7 ;
Z1 = c,
Z2 = b,
Z3 = h(b,c),
Deg = 0.6 ;
Z1 = c,
Z2 = b,
Z3 = h(b,b),
Deg = 0.6 ;
false.

?- all_instances(f(h(Z1,Z2),Z1,Z2,Z3), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, I).
I = [[f(h(a,b),a,b,h(a,b)),0.6],[f(h(a,b),a,b,h(a,c)),0.6],[f(h(a,b),a,b,h(b,b)),0.6],[f(h(a,b),a,b,h(b,c)),0.6],[f(h(c,b),c,b,h(a,b)),0.7],[f(h(c,b),c,b,h(a,c)),0.7],[f(h(c,b),c,b,h(b,b)),0.6],[f(h(c,b),c,b,h(b,c)),0.6]].

?- best_instances(f(h(Z1,Z2),Z1,Z2,Z3), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, I).
I = [[f(h(c,b),c,b,h(a,c)),0.7],[f(h(c,b),c,b,h(a,b)),0.7]].

?- is_proximal(a, b, 0.7).
false.

?- is_proximal(a, b, 0.4).
true.

?- is_proximal(h(X,X), h(a,c), Deg).
Deg = 0.

?- is_proximal(_, _, 0).
true.

?- is_proximal(_, _, 0.1).
false.

?- is_proximal(X, X, 0.1).
X = c11.

?- is_proximal(h(X,X), h(a,c), 0).
true.

?- is_proximal(h(X,X), h(a,c), 0.1).
false.

?- is_proximal(h(X,X), h(X,X), Deg).
X = c28,
Deg = 1.



*****************************************/

% Unit tests.
% To test, evaluate run_tests(rel1).

% Note that tests solve3, solve4, and solve6 look like 
% unification problems, but they are in a form that 
% can be dealt with as if they were matching problems.


:- begin_tests(rel1).

test(solve1) :-
    findall([X,Deg], solve(X, a, 0, Deg), L),
    L == [[a,1],[b,0.6]].

test(solve2) :-
    findall([X,Deg], solve(X, a, 0.8, Deg), L),
    L == [[a,1]].

test(solve3) :-
    findall([X,Deg], solve(b, X, 0, Deg), L),
    L == [[b,1],[a,0.6],[c,0.7]].

test(solve4) :-
    findall([X,Deg], solve(b, X, 0.7, Deg), L),
    L == [[b,1],[c,0.7]].

test(solve5) :-
    findall([X,Deg], solve(h(X,X), h(a,c), 0, Deg), L),
    L == [[b,0.6]].

test(solve6) :-
    findall([X,Deg], solve(h(X,c), h(a,X), 0, Deg), L),
    L == [[b,0.6]].

test(solve7) :-
    findall([X,Deg], solve(h(X,X), h(b,b), 0, Deg), L),
    L == [[b,1],[a,0.6],[c,0.7]].

test(solve8) :-
    findall([X,Deg], solve(h(X,X), h(b,b), 0.7, Deg), L),
    L == [[b,1],[c,0.7]].

test(solve9) :-
    set_prolog_flag(answer_write_options,[max_depth(0)]),
    findall([Z1,Z2,Z3,Deg], solve(f(h(Z1,Z2),Z1,Z2,Z3), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, Deg), L),
    L = [[a,b,h(a,c),0.6],[a,b,h(a,b),0.6],[a,b,h(b,c),0.6],[a,b,h(b,b),0.6],[c,b,h(a,c),0.7],[c,b,h(a,b),0.7],[c,b,h(b,c),0.6],[c,b,h(b,b),0.6]].

test(all_instances1) :-
    set_prolog_flag(answer_write_options,[max_depth(0)]),
    all_instances(f(h(Z1,Z2),Z1,Z2,_), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, I),
    I = [[f(h(a,b),a,b,h(a,b)),0.6],[f(h(a,b),a,b,h(a,c)),0.6],[f(h(a,b),a,b,h(b,b)),0.6],[f(h(a,b),a,b,h(b,c)),0.6],[f(h(c,b),c,b,h(a,b)),0.7],[f(h(c,b),c,b,h(a,c)),0.7],[f(h(c,b),c,b,h(b,b)),0.6],[f(h(c,b),c,b,h(b,c)),0.6]].

test(best_instances1) :-
    set_prolog_flag(answer_write_options,[max_depth(0)]),
    best_instances(f(h(Z1,Z2),Z1,Z2,_), g(b,h(a,c),a,h(a,b),h(a,c)), 0.4, I),
    I = [[f(h(c,b),c,b,h(a,c)),0.7],[f(h(c,b),c,b,h(a,b)),0.7]].

test(is_proximal1) :-
    \+ is_proximal(a, b, 0.7).

test(is_proximal2) :-
    is_proximal(a, b, 0.4).

test(is_proximal3) :-
    is_proximal(h(X,X), h(a,c), Deg),
    Deg == 0.

test(is_proximal4) :-
    is_proximal(_, _, 0).

test(is_proximal5) :-
    \+ is_proximal(_, _, 0.1).

test(is_proximal6) :-
    is_proximal(X, X, 0.1).

test(is_proximal7) :-
    is_proximal(h(X,X), h(a,c), 0).

test(is_proximal8) :-
    \+ is_proximal(h(X,X), h(a,c), 0.1).

test(is_proximal9) :-
    is_proximal(h(X,X), h(X,X), Deg),
    Deg == 1.

:- end_tests(rel1).
