/***************************************************

 These clauses define a sample proximity relation.
 
 Relation R3, 2023-03-23.
 Last version: 2023-04-30

 Example 2 from the FUZZ-IEEE 2021 paper:

 Cleo Pau, Temur Kutsia. Proximity-Based Unification 
 and Matching for Fully Fuzzy Signatures. 
 Proceedings of FUZZ-IEEE 2021. IEEE 2021.  
 DOI: 10.1109/FUZZ45933.2021.9494438

***************************************************/


:-include('generic.pl').

% A proximity equation between two nonconstant terms.
% Both symmetric versions should be written,
% unless the symmetric version causes looping.

prox(p(X), q(Y1,Y2), Delta) :-
  % defining constraint  
    prox(X, Y1, Delta1),
    prox(X, Y2, Delta2),
    min_list([0.7,Delta1,Delta2], Delta).
prox(q(Y1,Y2), p(X), Delta) :-
    prox(p(X), q(Y1,Y2), Delta).

prox(f(X1,_,_), g(Y), Delta) :-
  % defining constraint   
    prox(X1, Y, Delta1),
    min_list([0.6,Delta1], Delta).
prox(g(Y), f(X1,X2,X3), Delta) :-
    prox(f(X1,X2,X3), g(Y), Delta).

prox(f(_,_,X3), h(Y), Delta) :-
  % defining constraint   
    prox(X3, Y, Delta1),
    min_list([0.5,Delta1], Delta).
prox(h(Y), f(X1,X2,X3), Delta) :-
    prox(f(X1,X2,X3), h(Y), Delta).



% Proximity equations between different constants.
% Both symmetric versions should be written.

prox(b, c, 0.4).
prox(c, b, 0.4).



/*****************************************

Sample goals:

?- solve(p(X), q(g(a), h(c)), 0.4, Deg).
X = f(a,_,b),
Deg = 0.4 ;
X = f(a,_,c),
Deg = 0.5 ;
false.

?- all_instances(p(X), q(g(a), h(c)), 0.4, I).
I = [[p(f(a,_,b)),0.4],[p(f(a,_,c)),0.5]].

?- best_instances(p(X), q(g(a), h(c)), 0.4, I).
I = [[p(f(a,_,c)),0.5]].

*****************************************/

%  Unit tests
%  To test, evaluate run_tests(rel3).

:- begin_tests(rel3).

test(solve1) :-
    findall([X,Deg], solve(p(X), q(g(a), h(c)), 0.4, Deg), L),
    L = [[f(a,_,c),0.5],[f(a,_,b),0.4]].

test(all_instances1) :-
    all_instances(p(_), q(g(a), h(c)), 0.4, I),
    I = [[p(f(a,_,b)),0.4],[p(f(a,_,c)),0.5]].

test(best_instances1) :-
    best_instances(p(_), q(g(a), h(c)), 0.4, I),
    I = [[p(f(a,_,c)),0.5]].

:- end_tests(rel3).
