/***************************************************

 These clauses define a sample proximity relation.
 
 Relation R5, 2023-03-23.
 Last version: 2023-03-26

 Examples with infinite loops.
 They illustrate that when the conditions of the
 definition of conditional proximity are violated,
 then matching might not terminate.

***************************************************/

:-include('generic.pl').

% A proximity equation between two nonconstant terms.
% Both symmetric versions should be written,
% unless the symmetric version causes looping.


% This causes an infinite loop.
% Reason: restricting constraint comes before the
% defining constraint. Not all necessary variables
% are instantiated.
% Try solve(f(X1,X2), g(f(a)), 0.3, Deg).

prox(f(X1,X2), g(Y), Delta) :-
  % restricting and degree constraints 
    prox(X1, h(X2), Alpha),
    Alpha > 0,
    prox(X2, h(X1), Beta),
    Beta > 0,
  % defining constraint
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g(Y), f(X1,X2), Delta) :-
    prox(f(X1,X2), g(Y), Delta).

% This causes an infinite loop.
% Reason: restricting constraint comes before the
% defining constraint. Not all necessary variables
% are instantiated.
% Try: solve(f1(X1,X2), g1(a), 0.3, Deg).

prox(f1(X1,X2), g1(Y), Delta) :-
   % restricting and degree constraints   
    prox(X1, g1(X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g1(Y), f1(X1,X2), Delta) :-
    prox(f1(X1,X2), g1(Y), Delta).


% This causes an infinite loop.
% Reason: restricting constraint comes before the
% defining constraint. Not all necessary variables
% are instantiated.
% Try: solve(f2(X1,X2), g2(a), 0.3, Deg).

prox(f2(X1,X2), g2(Y), Delta) :-
  % restricting and degree constraints    
    prox(X1, g3(X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g2(Y), f2(X1,X2), Delta) :-
    prox(f2(X1,X2), g2(Y), Delta).

prox(f3(X1,X2), g3(Y), Delta) :-
  % restricting and degree constraints    
    prox(X1, g2(X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g3(Y), f3(X1,X2), Delta) :-
    prox(f3(X1,X2), g3(Y), Delta).


% This causes an infinite loop.
% Reason: restricting constraint comes before the
% defining constraint. Not all necessary variables
% are instantiated.
% Try: solve(f4(X1,X2), g4(a), 0.3, Deg).

prox(f4(X1,X2), g4(Y), Delta) :-
  % restricting and degree constraints    
    prox(X1, f4(X2,X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g4(Y), f4(X1,X2), Delta) :-
    prox(f4(X1,X2), g4(Y), Delta).


% This causes an infinite loop.
% Reason: restricting constraint comes before the
% defining constraint. Not all necessary variables
% are instantiated.
% Try: solve(f5(X1,X2), g5(a), 0.3, Deg).

prox(f5(X1,X2), g5(Y), Delta) :-
  % restricting and degree constraints    
    prox(X1, f6(Y,X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g5(Y), f5(X1,X2), Delta) :-
    prox(f5(X1,X2), g5(Y), Delta).

prox(f6(X1,X2), g6(Y), Delta) :-
  % restricting constraint    
    prox(X1, f5(Y,X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g6(Y), f6(X1,X2), Delta) :-
    prox(f6(X1,X2), g6(Y), Delta).


% This causes an infinite loop.
% Reason: in restricting constraint, the variable
% occurrence condition is violated: X3 does not
% occur in the definiting part.
% Try: solve(f7(X1,X2,X3), g7(a), 0.3, Deg).

prox(f7(X1,X2,X3), g7(Y), Delta) :-
  % defining constraint   
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta),
  % restricting and degree constraints    
    prox(X2, h(X3), Alpha),
    Alpha > 0.
prox(g7(Y), f7(X1,X2,X3), Delta) :-
    prox(f7(X1,X2,X3), g7(Y), Delta).
  


% Proximity equations between different constants.
% Both symmetric versions should be written.

prox(a, b, 0.5).
prox(b, a, 0.5).

prox(a, c, 0.7).
prox(c, a, 0.7).

prox(b, d, 0.7).
prox(d, b, 0.7).

prox(c, d, 0.8).
prox(d, c, 0.8).




/*****************************************

Sample goals, leading to infinite loops:

?- solve(f(X1,X2), g(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.7Gb, global: 0.3Gb, trail: 32.2Mb
ERROR:   Stack depth: 4,215,546, last-call: 0%, Choice points: 2,107,771
ERROR:   Possible non-terminating recursion:
ERROR:     [4,215,546] user:prox(_67449920, <compound h/1>, _67449924)
ERROR:     [4,215,545] user:prox(<compound f/2>, <compound g/1>, _67449952)
   Exception: (4,215,545) prox(_67449836,_67449812,_67449838) ? a
% Execution Aborted

?- solve(f1(X1,X2), g1(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.8Gb, global: 0.2Gb, trail: 31.0Mb
ERROR:   Stack depth: 4,066,387, last-call: 0%, Choice points: 4,066,379
ERROR:   In:
ERROR:     [4,066,387] user:prox(_48797894, <compound g1/1>, _48797898)
ERROR:     [4,066,386] user:prox(<compound f1/2>, <compound g1/1>, _48797926)
ERROR:     [4,066,385] user:prox(<compound f1/2>, <compound g1/1>, _48797960)
ERROR:     [4,066,384] user:prox(<compound f1/2>, <compound g1/1>, _48797994)
ERROR:     [4,066,383] user:prox(<compound f1/2>, <compound g1/1>, _48798028)
ERROR: 
ERROR: Use the --stack_limit=size[KMG] command line option or
ERROR: ?- set_prolog_flag(stack_limit, 2_147_483_648). to double the limit.
   Exception: (4,066,386) prox(_48797790,g1(_48797792),_48797798) ? a
% Execution Aborted



?- solve(f2(X1,X2), g2(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.8Gb, global: 0.2Gb, trail: 31.0Mb
ERROR:   Stack depth: 4,066,387, last-call: 0%, Choice points: 4,066,379
ERROR:   In:
ERROR:     [4,066,387] user:prox(_48797894, <compound g2/1>, _48797898)
ERROR:     [4,066,386] user:prox(<compound f3/2>, <compound g3/1>, _48797926)
ERROR:     [4,066,385] user:prox(<compound f2/2>, <compound g2/1>, _48797960)
ERROR:     [4,066,384] user:prox(<compound f3/2>, <compound g3/1>, _48797994)
ERROR:     [4,066,383] user:prox(<compound f2/2>, <compound g2/1>, _48798028)
ERROR: 
ERROR: Use the --stack_limit=size[KMG] command line option or
ERROR: ?- set_prolog_flag(stack_limit, 2_147_483_648). to double the limit.
   Exception: (4,066,386) prox(_48797790,g3(_48797792),_48797798) ? a 
   Exception: (4,066,385) prox(_48797778,g2(_48797780),_48797786) ? a
% Execution Aborted



?- solve(f3(X1,X2), g3(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.8Gb, global: 0.2Gb, trail: 31.0Mb
ERROR:   Stack depth: 4,066,387, last-call: 0%, Choice points: 4,066,379
ERROR:   In:
ERROR:     [4,066,387] user:prox(_48797894, <compound g3/1>, _48797898)
ERROR:     [4,066,386] user:prox(<compound f2/2>, <compound g2/1>, _48797926)
ERROR:     [4,066,385] user:prox(<compound f3/2>, <compound g3/1>, _48797960)
ERROR:     [4,066,384] user:prox(<compound f2/2>, <compound g2/1>, _48797994)
ERROR:     [4,066,383] user:prox(<compound f3/2>, <compound g3/1>, _48798028)
ERROR: 
ERROR: Use the --stack_limit=size[KMG] command line option or
ERROR: ?- set_prolog_flag(stack_limit, 2_147_483_648). to double the limit.
   Exception: (4,066,386) prox(_48797790,g2(_48797792),_48797798) ? a
% Execution Aborted


?- solve(f4(X1,X2), g4(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.8Gb, global: 0.2Gb, trail: 0Kb
ERROR:   Stack depth: 4,587,585, last-call: 0%, Choice points: 4,587,577
ERROR:   Possible non-terminating recursion:
ERROR:     [4,587,585] user:prox(<compound f4/2>, <compound g4/1>, _41289538)
ERROR:     [4,587,584] user:prox(<compound g4/1>, <compound f4/2>, _41289572)
   Exception: (4,587,584) prox(g4(_1310),f4(g4(_1310),_864),_41289458) ? a
% Execution Aborted



?- solve(f5(X1,X2), g5(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.7Gb, global: 0.3Gb, trail: 32.2Mb
ERROR:   Stack depth: 4,215,473, last-call: 0%, Choice points: 2,107,735
ERROR:   In:
ERROR:     [4,215,473] user:prox(_67452114, <compound h/1>, _67452118)
ERROR:     [4,215,472] user:prox(<compound f/2>, <compound g/1>, _67452146)
ERROR:     [4,215,471] user:prox(_67452176, <compound h/1>, _67452180)
ERROR:     [4,215,470] user:prox(<compound f/2>, <compound g/1>, _67452208)
ERROR:     [4,215,469] user:prox(_67452238, <compound h/1>, _67452242)
ERROR: 
ERROR: Use the --stack_limit=size[KMG] command line option or
ERROR: ?- set_prolog_flag(stack_limit, 2_147_483_648). to double the limit.
   Exception: (4,215,472) prox(_67452012,_67451988,_67452014) ? a
% Execution Aborted



?- solve(f7(X1,X2,X3), g7(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.7Gb, global: 0.3Gb, trail: 32.2Mb
ERROR:   Stack depth: 4,215,354, last-call: 0%, Choice points: 2,107,676
ERROR:   Possible non-terminating recursion:
ERROR:     [4,215,354] user:prox(_67454714, <compound h/1>, _67454718)
ERROR:     [4,215,353] user:prox(<compound f/2>, <compound g/1>, _67454746)
   Exception: (4,215,353) prox(_67454630,_67454606,_67454632) ? a
% Execution Aborted

*****************************************/

