/***************************************************

 These clauses define a sample proximity relation.
 
 Relation R6, 2023-03-23.
 Last version: 2023-04-30

 Examples from prox-rel-05.pl, 
 modified to avoid infinite loops.

***************************************************/

:-include('generic.pl').

% A proximity equation between two nonconstant terms.
% Both symmetric versions should be written,
% unless the symmetric version causes looping.


% Bringing prox(X1, Y, Delta1) before the part
% that caused an infinite loop (i.e., bringing
% the defining constraint before the restricting
% constraint) helps to get necessary instatiation of
% variables. 
%
% Try solve(f(X1,X2), g(f(a)), 0.3, Deg).
% No loop anymore. Now it fails.

prox(f(X1,X2), g(Y), Delta) :-
  % defining constraint  
    prox(X1, Y, Delta1),
  % restricting constraint  
    prox(X1, h(X2), Alpha),
    Alpha > 0,
    prox(X2, h(X1), Beta),
    Beta > 0,
    min_list([0.7,Delta1], Delta).
prox(g(Y), f(X1,X2), Delta) :-
    prox(f(X1,X2), g(Y), Delta).


% Bringing prox(X1, Y, Delta1) before the part
% that caused an infinite loop (i.e., bringing
% the defining constraint before the restricting
% constraint) helps to get necessary instatiation of
% variables. 
%
% Try: solve(f1(X1,X2), g1(a), 0.3, Deg).
% No loop anymore. Now it fails.

prox(f1(X1,X2), g1(Y), Delta) :-
  % defining constraint    
    prox(X1, Y, Delta1),
  % restricting constraint  
    prox(X1, g1(X2), Alpha),
    Alpha > 0,
    min_list([0.7,Delta1], Delta).
prox(g1(Y), f1(X1,X2), Delta) :-
    prox(f1(X1,X2), g1(Y), Delta).


% Bringing prox(X1, Y, Delta1) before the part
% that caused an infinite loop (i.e., bringing
% the defining constraint before the restricting
% constraint) helps to get necessary instatiation of
% variables. 
%
% Try: solve(f2(X1,X2), g2(a), 0.3, Deg).
% No loop anymore. Now it fails.
%
% But solve(f3(X1,X2), g3(a), 0.3, Deg) still loops.
% If we swap the defining and restricging constraints
% in the clause with the head prox(f3(X1,X2), g3(Y), Delta),
% it would cause solve(f3(X1,X2), g3(a), 0.3, Deg) to fail.

prox(f2(X1,X2), g2(Y), Delta) :-
  % defining constraint    
    prox(X1, Y, Delta1),
  % restricting constraint   
    prox(X1, g3(X2), Alpha),
    Alpha > 0,
    min_list([0.7,Delta1], Delta).
prox(g2(Y), f2(X1,X2), Delta) :-
    prox(f2(X1,X2), g2(Y), Delta).

prox(f3(X1,X2), g3(Y), Delta) :-
  % restricting constraint   
    prox(X1, g2(X2), Alpha),
    Alpha > 0,
  % defining constraint  
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g3(Y), f3(X1,X2), Delta) :-
    prox(f3(X1,X2), g3(Y), Delta).

% Bringing prox(X1, Y, Delta1) before the part
% that caused an infinite loop (i.e., bringing
% the defining constraint before the restricting
% constraint) helps to get necessary instatiation of
% variables. 
%
% Try: solve(f4(X1,X2), g4(a), 0.3, Deg).
% No loop anymore. Now it fails.

prox(f4(X1,X2), g4(Y), Delta) :-
  % defining constraint    
    prox(X1, Y, Delta1),
  % restricting constraint   
    prox(X1, f4(X2,X2), Alpha),
    Alpha > 0,
    min_list([0.7,Delta1], Delta).
prox(g4(Y), f4(X1,X2), Delta) :-
    prox(f4(X1,X2), g4(Y), Delta).


% Bringing prox(X1, Y, Delta1) before the part
% that caused an infinite loop (i.e., bringing
% the defining constraint before the restricting
% constraint) helps to get necessary instatiation of
% variables. 
%
% Try: solve(f5(X1,X2), g5(a), 0.3, Deg).
% No loop anymore. Now it fails.
%
% But solve(f6(X1,X2), g6(a), 0.3, Deg) still loops.
% If we swap the defining and restricging constraints  
% in the clause with the head prox(f3(X1,X2), g3(Y), Delta), 
% it would cause solve(f6(X1,X2), g6(a), 0.3, Deg) to fail.

prox(f5(X1,X2), g5(Y), Delta) :-
  % defining constraint    
    prox(X1, Y, Delta1),
  % restricting constraint   
    prox(X1, f6(Y,X2), Alpha),
    Alpha > 0,
    min_list([0.7,Delta1], Delta).
prox(g5(Y), f5(X1,X2), Delta) :-
    prox(f5(X1,X2), g5(Y), Delta).

prox(f6(X1,X2), g6(Y), Delta) :-
  % restricting constraint    
    prox(X1, f5(Y,X2), Alpha),
    Alpha > 0,
  % defining constraint   
    prox(X1, Y, Delta1),
    min_list([0.7,Delta1], Delta).
prox(g6(Y), f6(X1,X2), Delta) :-
    prox(f6(X1,X2), g6(Y), Delta).

% Unlike its counterpart from prox-rel-5.pl,
% here X3 appears in the defining constraint.
%
% Try: solve(f7(X1,X2,X3), g7(a), 0.3, Deg).

prox(f7(X1,X2,X3), g7(Y), Delta) :-
  % defining constraint   
    prox(X1, Y, Delta1),
    prox(X3, Y, Delta2),
    min_list([0.7,Delta1,Delta2], Delta),
  % restricting constraint    
    prox(X2, h(X3), Alpha),
    Alpha > 0.
prox(g7(Y), f7(X1,X2,X3), Delta) :-
    prox(f7(X1,X2,X3), g7(Y), Delta).

% Proximity equations between different constants.
% Both symmetric versions should be written.

prox(a, b, 0.5).
prox(b, a, 0.5).

prox(a, c, 0.7).
prox(c, a, 0.7).

prox(b, d, 0.7).
prox(d, b, 0.7).

prox(c, d, 0.8).
prox(d, c, 0.8).




/*****************************************

Sample goals, 

?- solve(f(X1,X2), g(a), 0.3, Deg).
false.

?- solve(f1(X1,X2), g1(a), 0.3, Deg).
false.

?- solve(f2(X1,X2), g2(a), 0.3, Deg).
false.

?- solve(f3(X1,X2), g3(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.7Gb, global: 0.2Gb, trail: 58.5Mb
ERROR:   Stack depth: 3,833,897, last-call: 0%, Choice points: 3,833,889
ERROR:   Possible non-terminating recursion:
ERROR:     [3,833,897] user:prox(_46013786, _46013788, _46013790)
ERROR:     [3,833,896] user:prox(<compound f/2>, <compound g/1>, _46013812)
   Exception: (3,833,896) prox(_46013700,_46013706,_46013708) ? a
% Execution Aborted
?- solve(f4(X1,X2), g4(a), 0.3, Deg).
false.

?- solve(f5(X1,X2), g5(a), 0.3, Deg).
false.

?- solve(f6(X1,X2), g6(a), 0.3, Deg).
ERROR: Stack limit (1.0Gb) exceeded
ERROR:   Stack sizes: local: 0.7Gb, global: 0.2Gb, trail: 58.5Mb
ERROR:   Stack depth: 3,833,976, last-call: 0%, Choice points: 3,833,967
ERROR:   Possible non-terminating recursion:
ERROR:     [3,833,976] user:prox(_46011438, _46011440, _46011442)
ERROR:     [3,833,975] user:prox(<compound f/2>, <compound g/1>, _46011464)
   Exception: (3,833,975) prox(_46011352,_46011358,_46011360) ? a
% Execution Aborted


?- solve(f7(X1,X2,X3), g7(a), 0.3, Deg).
X1 = X3, X3 = a,
X2 = h(a),
Deg = 0.7 ;
X1 = X3, X3 = a,
X2 = h(b),
Deg = 0.7 ;
X1 = X3, X3 = a,
X2 = h(c),
Deg = 0.7 ;
X1 = a,
X2 = h(b),
X3 = b,
Deg = 0.5 ;
X1 = a,
X2 = h(a),
X3 = b,
Deg = 0.5 ;
X1 = a,
X2 = h(d),
X3 = b,
Deg = 0.5 ;
X1 = a,
X2 = h(c),
X3 = c,
Deg = 0.7 ;
X1 = a,
X2 = h(a),
X3 = c,
Deg = 0.7 ;
X1 = a,
X2 = h(d),
X3 = c,
Deg = 0.7 ;
X1 = b,
X2 = h(a),
X3 = a,
Deg = 0.5 ;
X1 = b,
X2 = h(b),
X3 = a,
Deg = 0.5 ;
X1 = b,
X2 = h(c),
X3 = a,
Deg = 0.5 ;
X1 = X3, X3 = b,
X2 = h(b),
Deg = 0.5 ;
X1 = X3, X3 = b,
X2 = h(a),
Deg = 0.5 ;
X1 = X3, X3 = b,
X2 = h(d),
Deg = 0.5 ;
X1 = b,
X2 = h(c),
X3 = c,
Deg = 0.5 ;
X1 = b,
X2 = h(a),
X3 = c,
Deg = 0.5 ;
X1 = b,
X2 = h(d),
X3 = c,
Deg = 0.5 ;
X1 = c,
X2 = h(a),
X3 = a,
Deg = 0.7 ;
X1 = c,
X2 = h(b),
X3 = a,
Deg = 0.7 ;
X1 = c,
X2 = h(c),
X3 = a,
Deg = 0.7 ;
X1 = c,
X2 = h(b),
X3 = b,
Deg = 0.5 ;
X1 = c,
X2 = h(a),
X3 = b,
Deg = 0.5 ;
X1 = c,
X2 = h(d),
X3 = b,
Deg = 0.5 ;
X1 = X3, X3 = c,
X2 = h(c),
Deg = 0.7 ;
X1 = X3, X3 = c,
X2 = h(a),
Deg = 0.7 ;
X1 = X3, X3 = c,
X2 = h(d),
Deg = 0.7 ;
false.


*****************************************/

% Unit tests.
% To test, evaluate run_tests(rel6).

:- begin_tests(rel6).

test(solve1) :-
    set_prolog_flag(answer_write_options,[max_depth(0)]),
    findall([X1,X2,X3,Deg], solve(f7(X1,X2,X3), g7(a), 0.3, Deg), L),
    L == [[a,h(a),a,0.7],[a,h(b),a,0.7],[a,h(c),a,0.7],[a,h(b),b,0.5],[a,h(a),b,0.5],[a,h(d),b,0.5],[a,h(c),c,0.7],[a,h(a),c,0.7],[a,h(d),c,0.7],[b,h(a),a,0.5],[b,h(b),a,0.5],[b,h(c),a,0.5],[b,h(b),b,0.5],[b,h(a),b,0.5],[b,h(d),b,0.5],[b,h(c),c,0.5],[b,h(a),c,0.5],[b,h(d),c,0.5],[c,h(a),a,0.7],[c,h(b),a,0.7],[c,h(c),a,0.7],[c,h(b),b,0.5],[c,h(a),b,0.5],[c,h(d),b,0.5],[c,h(c),c,0.7],[c,h(a),c,0.7],[c,h(d),c,0.7]].

test(all_instances1) :-
    all_instances(f7(_,_,_), g7(a), 0.3, I),
    I = [[f7(a,h(a),a),0.7],[f7(a,h(a),b),0.5],[f7(a,h(a),c),0.7],[f7(a,h(b),a),0.7],[f7(a,h(b),b),0.5],[f7(a,h(c),a),0.7],[f7(a,h(c),c),0.7],[f7(a,h(d),b),0.5],[f7(a,h(d),c),0.7],[f7(b,h(a),a),0.5],[f7(b,h(a),b),0.5],[f7(b,h(a),c),0.5],[f7(b,h(b),a),0.5],[f7(b,h(b),b),0.5],[f7(b,h(c),a),0.5],[f7(b,h(c),c),0.5],[f7(b,h(d),b),0.5],[f7(b,h(d),c),0.5],[f7(c,h(a),a),0.7],[f7(c,h(a),b),0.5],[f7(c,h(a),c),0.7],[f7(c,h(b),a),0.7],[f7(c,h(b),b),0.5],[f7(c,h(c),a),0.7],[f7(c,h(c),c),0.7],[f7(c,h(d),b),0.5],[f7(c,h(d),c),0.7]].

test(best_instances1) :-
    best_instances(f7(_,_,_), g7(a), 0.3, I),
    I = [[f7(a,h(a),a),0.7],[f7(a,h(b),a),0.7],[f7(a,h(c),a),0.7],[f7(a,h(c),c),0.7],[f7(a,h(a),c),0.7],[f7(a,h(d),c),0.7],[f7(c,h(a),a),0.7],[f7(c,h(b),a),0.7],[f7(c,h(c),a),0.7],[f7(c,h(c),c),0.7],[f7(c,h(a),c),0.7],[f7(c,h(d),c),0.7]].


:- end_tests(rel6).
